/**
 * @file ISAM2_SmartFactorStereo_IMU.cpp
 * @brief test of iSAM2 with smart stereo factors and IMU preintegration,
 * originally used to debug valgrind invalid reads with Eigen
 * @author Nghia Ho
 *
 * Setup is a stationary stereo camera with an IMU attached.
 * The data file is at examples/Data/ISAM2_SmartFactorStereo_IMU.txt
 * It contains 5 frames of stereo matches and IMU data.
 */
#include <ros/ros.h>
#include <sensor_msgs/Imu.h>
#include <sensor_msgs/PointCloud2.h>
#include <geometry_msgs/PoseStamped.h>
#include <geometry_msgs/PointStamped.h>

#include <gtsam/navigation/CombinedImuFactor.h>
#include <gtsam/nonlinear/ISAM2.h>
#include <gtsam_unstable/slam/SmartStereoProjectionPoseFactor.h>
#include <gtsam/navigation/ImuFactor.h>
#include <gtsam/sam/BearingRangeFactor.h>
#include <gtsam/navigation/GPSFactor.h>
#include <gtsam/navigation/AttitudeFactor.h>
#include <gtsam/slam/PoseRotationPrior.h>

// #include <gtsam_unstable/dynamics/VelocityConstraint.h>
// #include <gtsam/nonlinear/NonlinearFactor.h>
// #include <gtsam/nonlinear/Marginals.h>

// #include <pcl_conversions/pcl_conversions.h>
// #include <pcl/point_cloud.h>
// #include <pcl/point_types.h>

#include "isam2/orientationEstimate.h"
#include "isam2/NorthEastHeading.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

using namespace std;
using namespace gtsam;
using symbol_shorthand::X;
using symbol_shorthand::V;
using symbol_shorthand::B;
using symbol_shorthand::L;

ros::Publisher pub;
int pose_counter=0;
int landmark_counter=0;
int number_landmarks = 0;
Eigen::Matrix3f total_rotation;
Eigen::Matrix3f rotation;
Eigen::Vector3f total_translation;
Eigen::Vector3f translation;
Eigen::Quaternionf q;
Pose3 total_pose2;
ofstream myfile;
Rot3 orientation;

ros::Publisher pub2;

ISAM2Params parameters;
ISAM2 isam;

void do_more();

// Create a factor graph
// std::map<size_t, SmartStereoProjectionPoseFactor::shared_ptr> smartFactors;
NonlinearFactorGraph graph;
Values initialEstimate;

int lastFrame = 1;
int counter = 1;
int imu_counter = 1;
int gps_pose_counter = 0;

double north = 0.0;
double east = 0.0;
double theta = 0.0;
double theta_orientation = 0.0;
bool nav_online = false;
double prev_east = 0.0;
double prev_north = 0.0;
double prev_theta = 0.0;
bool first_landmark = true;

vector<int> ids;

ofstream file;

struct IMUHelper {
  IMUHelper() {
    {
      auto gaussian = noiseModel::Diagonal::Sigmas(
          (Vector(6) << Vector3(0.15, 0.15, 0.15), Vector3(0.15, 0.15, 0.15))
              .finished());
      auto huber = noiseModel::Robust::Create(
          noiseModel::mEstimator::Huber::Create(1.345), gaussian);

      biasNoiseModel = huber;
    }

    {
      auto gaussian = noiseModel::Isotropic::Sigma(3, 0.3);
      auto huber = noiseModel::Robust::Create(
          noiseModel::mEstimator::Huber::Create(1.345), gaussian);

      velocityNoiseModel = huber;
    }

    // expect IMU to be rotated in image space co-ords
    auto p = boost::make_shared<PreintegratedCombinedMeasurements::Params>(
        Vector3(0.0, 0.0, 9.81));

    p->accelerometerCovariance =
        I_3x3 * 0.004;  // acc white noise in continuous
    p->integrationCovariance =
        I_3x3 * 0.002;  // integration uncertainty continuous
    p->gyroscopeCovariance =
        I_3x3 * 0.001;  // gyro white noise in continuous
    p->biasAccCovariance = I_3x3 * 0.004;  // acc bias in continuous
    p->biasOmegaCovariance =
        I_3x3 * 0.001;  // gyro bias in continuous
    p->biasAccOmegaInt = Matrix::Identity(6, 6) * 1e-4;

    Rot3 pitch = Rot3::pitch(0.0);
    Rot3 roll = Rot3::roll(0.0);
    // Rot3 yaw = Rot3::yaw(-79);
    // Rot3 yaw = Rot3::yaw(1);

    // body to IMU rotation
    Rot3 iRb = pitch * roll;

    // body to IMU translation (meters)
    Point3 iTb(0, 0, 3);

    // body in this example is the left camera
    p->body_P_sensor = Pose3(iRb, iTb);

    // Rot3 prior_rotation = Rot3::yaw(0.83);
    Rot3 prior_rotation = Rot3(I_3x3);
    Pose3 prior_pose(prior_rotation, Point3(0, 0, 0));
    // Pose3 prior_pose(prior_rotation, Point3(-0.645953733353, 2.79517319837, 0));

    Vector3 acc_bias(0.2, -0.2, -0.04);  // in camera frame
    Vector3 gyro_bias(0.005, 0.0006, 0.024);

    priorImuBias = imuBias::ConstantBias(acc_bias, gyro_bias);

    prevState = NavState(prior_pose, Vector3(0, 0, 0));
    propState = prevState;
    prevBias = priorImuBias;

    preintegrated = new PreintegratedCombinedMeasurements(p, priorImuBias);
  }

  imuBias::ConstantBias priorImuBias;  // assume zero initial bias
  noiseModel::Robust::shared_ptr velocityNoiseModel;
  noiseModel::Robust::shared_ptr biasNoiseModel;
  NavState prevState;
  NavState propState;
  imuBias::ConstantBias prevBias;
  PreintegratedCombinedMeasurements* preintegrated;
};

IMUHelper imu;

void point_test(const geometry_msgs::PointStamped point_msg){
  // std::cout << "Landmark= " << landmark_counter << ", Pose= " << pose_counter << std::endl;
  std::stringstream counter(point_msg.header.frame_id);
  counter >> landmark_counter;
  Point3 point(point_msg.point.x, point_msg.point.y, point_msg.point.z);
  Eigen::Vector3f point_translation;
  point_translation << point_msg.point.x, point_msg.point.y, point_msg.point.z;
  Eigen::Matrix3f point_rotation;
  // point_rotation << total_rotation(0, 0), -total_rotation(0, 1), total_rotation(0, 2),  -total_rotation(1, 0),  total_rotation(1, 1),  total_rotation(1, 2),  total_rotation(2, 0),  total_rotation(2, 1),  total_rotation(2, 2);
  point_rotation << total_rotation(0, 0), -total_rotation(0, 1), total_rotation(0, 2),  -total_rotation(1, 0),  total_rotation(1, 1),  total_rotation(1, 2),  total_rotation(2, 0),  total_rotation(2, 1),  total_rotation(2, 2);
  // cout << "Before: " << point_msg.point.x << ", " << point_msg.point.y << ", " << point_msg.point.z << endl;
  Eigen::Vector3f current_translation = total_translation + point_rotation * point_translation;
  // cout << "Translation: " << total_translation << ", rotation: " << total_rotation << endl;
  // cout << "Total translation\n" << total_translation << endl;
  // cout << "Total rotation\n" << total_rotation << endl;
  // cout << "After transformation: " << landmark_counter << " " << current_translation(0) << ", " << current_translation(1) << ", " << current_translation(2) << endl;
  // Eigen::Vector3f current_translation = total_translation + point_translation;
  // Eigen::Vector3f current_translation = nav_trans + nav_rot * point_translation;
  Point3 total_point(current_translation(0), current_translation(1), current_translation(2));
  // Point3 total_point(point_msg.point.x+current_translation(0), point_msg.point.y+current_translation(1), point_msg.point.z+current_translation(2));
  // Point3 total_point(point_msg.point.x, point_msg.point.y, point_msg.point.z);
  bool exists = std::find(std::begin(ids), std::end(ids), landmark_counter) != std::end(ids);
  if (first_landmark){
    // total_point = Point3(0,0,0);
    static auto kPointPrior = noiseModel::Isotropic::Sigma(3, 0.3);
    graph.push_back(PriorFactor<Point3>(L(landmark_counter), total_point, kPointPrior));
    // BearingRangeFactor<Pose3, Point3> bFactor(X(pose_counter), L(landmark_counter), bearing, range, noise);
    // graph.push_back(BearingRangeFactor<Pose3, Point3>(bFactor));
    first_landmark = false;
  }
  // else{
  if (!exists){
  // if (!exists || ids.size() > landmark_counter + 200){
    // std::cout << point_msg.point.x << std::endl;
    // std::cout << total_pose2.bearing(point) << std::endl;
    // cout << "Total pose: \n" << total_pose2 << endl;
    // cout << "Total point: \n" << total_point << endl;
    BearingRange<Pose3,Point3> br;
    Unit3 bearing = br.MeasureBearing(total_pose2, total_point);
    // Rot3 bearing = Rot3::quaternion(0,2,2,0);
    double range = br.MeasureRange(total_pose2, total_point);
    noiseModel::Diagonal::shared_ptr noise = noiseModel::Diagonal::Sigmas(Vector3(0.3, 0.3, 0.3));
    // Vector3 sigmas;
    // sigmas << Vector3::Constant(0.1), Vector1::Constant(0.1);
    // auto noise = noiseModel::Diagonal::Sigmas(sigmas);
    // BearingRangeFactor<Pose3, Point3> bfactor;
    BearingRangeFactor<Pose3, Point3> bFactor(X(pose_counter), L(landmark_counter), bearing, range, noise);
    graph.push_back(BearingRangeFactor<Pose3, Point3>(bFactor));
  }
  // }
  
  if (!exists){
    initialEstimate.insert(Symbol('l', landmark_counter), total_point);
    ids.push_back(landmark_counter);
    if (landmark_counter>number_landmarks){
    number_landmarks = landmark_counter;      
    }
    // std::cout << "First time, " << "Landmark counter: " << landmark_counter << ", x: " << current_translation(0) << ", y: " << current_translation(1) << ", z: " << current_translation(2) << std::endl;
  } else {//if (ids.size() > landmark_counter + 1){
    std::cout << "Not first time, " << "Landmark counter: " << landmark_counter << ", x: " << current_translation(0) << ", y: " << current_translation(1) << ", z: " << current_translation(2) << std::endl;
  }

  // number_landmarks = std::max(ids);

  // if (!initialEstimate.exists(Symbol('l', landmark_counter))){
  //   std::cout << "Landmark does not exist: " << landmark_counter << std::endl;
  //   initialEstimate.insert(Symbol('l', landmark_counter), total_point);
  // }

  // landmark_counter++;
}

void pose_test(const geometry_msgs::PoseStamped pose_msg){

  if (pose_counter==0){
    // std::cout << "First" << std::endl;
    // if (nav_online){
      // Vector6 sigmas_; 
      // sigmas_ << Vector3(0.1, 0.1, 0.05), Vector3(0.1, 0.1, 0.5);
      // auto priorPoseNoise = noiseModel::Diagonal::Variances(sigmas_);
      // Pose3 initial(Rot3(0.677, -0.736, 0, 0.736, 0.677, 0, 0, 0, 1), Point3(-0.645953733353, 2.79517319837, 0));
      // graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
      // initialEstimate.insert(X(0), initial);
    // } else if (!nav_online){
      Vector6 sigmas_;
      sigmas_ << Vector3(0.1, 0.1, 0.1), Vector3(0.1, 0.1, 0.1);
      auto priorPoseNoise = noiseModel::Diagonal::Sigmas(sigmas_);

      if (false){
        file << north << ", " << east << ", " << 0 << endl;
        Pose3 initial(Rot3::Yaw(theta), Point3(north, east, 0));
        graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
        initialEstimate.insert(X(0), initial);
      } else {
        file << -0.65 << ", " << 2.8 << ", " << 0 << endl;
        Pose3 initial(Rot3(cos(0.83), sin(0.83), 0, -sin(0.83), cos(0.83), 0, 0, 0, 1), Point3(-0.645953733353, 2.79517319837, 0));
        graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
        initialEstimate.insert(X(0), initial);
      }
    // }
    graph.add(PriorFactor<imuBias::ConstantBias>(B(0), imu.priorImuBias,
                                               imu.biasNoiseModel));
    initialEstimate.insert(B(0), imu.priorImuBias);

    // Velocity prior - assume stationary
    graph.add(
        PriorFactor<Vector3>(V(0), Vector3(0, 0, 0), imu.velocityNoiseModel));
    initialEstimate.insert(V(0), Vector3(0, 0, 0));
    // prev_east = east;
    // prev_north = north;
    // prev_theta = theta;
  }
  std::stringstream counter(pose_msg.header.frame_id);
  counter >> pose_counter;
  // std::cout << "Pose: " << pose_msg.header.frame_id << std::endl;
  // std::cout << "Pose: " << pose_msg.pose << std::endl;
  // pose_counter++; 
  // Pose3 pose, delta;

  // if (nav_online && false){
    // double new_theta = theta-prev_theta;
    // double new_east = east;
    // double new_north = north;

    // // // cout << "East: " << new_east << ", prev_east: " << prev_east << ", north: " << new_north << ", prev_north: " << prev_north << endl;
    // double d2 = new_east-prev_east;
    // double d1 = new_north-prev_north;

    // // cout << "Change north: " << d1 << ", change east: " << d2 << ", change theta: " << new_theta << endl;

    // Rot3 prev_rot(Rot3::yaw(prev_theta));
    // Rot3 new_rot(Rot3::yaw(theta));

    // Point3 prev_tra(prev_north, prev_east, 0);
    // Point3 new_tra(north, east, 0);

    // Rot3 tot_rot = new_rot.between(prev_rot);
    // Point3 tot_tra = tot_rot * new_tra.between(prev_tra);

    // // cout << "Totrot: " << tot_rot << "\n Tottra: " << tot_tra << endl;

    // // Pose3 prev_pose(Rot3::yaw(theta), Point3(north, east, 0));
    // Pose3 prev_pose(Rot3(total_rotation(0,0), total_rotation(0,1), total_rotation(0,2), total_rotation(1,0), total_rotation(1,1), total_rotation(1,2), total_rotation(2,0), total_rotation(2,1), total_rotation(2,2)), Point3(total_translation(0), total_translation(1), total_translation(2)));
    // Pose3 new_pose(Rot3::yaw(theta), Point3(north, east, 0));
 
    // cout << "Total translation: " << total_translation(0) << " " << total_translation(1) << " " << total_translation(2) << endl;
    // cout << "Gps translation: " << north << " " << east << " " << "0" << endl;
    // // pose = Pose3(tot_rot, -tot_tra);

    // pose = new_pose.between(prev_pose);
    // pose = Pose3(Rot3::yaw(theta), Point3(north, east, 0));
    // sigmas << Vector3::Constant(0.01), Vector3::Constant(0.01);
    // Pose3 delta(Rot3::Rodrigues(0.01, 0.01, 0.01), Point3(0.01, 0.01, 0.01));
    // graph.push_back(PriorFactor<Pose3>(G(gps_pose_counter), pose, noise));

    // prev_east = east;
    // prev_north = north;
    // prev_theta = theta;

  // } else {
    Pose3 pose = Pose3(Rot3::Quaternion(pose_msg.pose.orientation.w, pose_msg.pose.orientation.x, pose_msg.pose.orientation.y, pose_msg.pose.orientation.z), Point3(pose_msg.pose.position.x, pose_msg.pose.position.y, pose_msg.pose.position.z));
    // cout << "Total rotation: \n" << total_rotation << endl;
    Vector6 sigmas;
    sigmas << Vector3(0.02, 0.02, 0.005), Vector3(0.01, 0.01, 0.01);
    // sigmas << Vector3(0.1, 0.1, 0.1), Vector3(0.1, 0.1, 0.1);
    Pose3 delta(Rot3::Rodrigues(0.1, 0.1, 0.03), Point3(0.035, 0.035, 0.02));
    // Pose3 delta(Rot3::Rodrigues(0.01, 0.01, 0.01), Point3(0.025, 0.025, 0.025));
    // Pose3 delta(Rot3::Rodrigues(0.7, 0.7, 0.7), Point3(0.2, 0.2, 0.2));
    auto noise = noiseModel::Diagonal::Sigmas(sigmas);
    graph.push_back(BetweenFactor<Pose3>(X(pose_counter-1), X(pose_counter), pose, noise));
  // }
  bool use_nav = false;
  if (nav_online && false && pose_counter < 230  /*&& false && pose_counter % 200 == 0 /*&& ( pose_counter < 30 || pose_counter > 90 && pose_counter < 110 )/*&& false /*&& ( pose_counter < 20 || (pose_counter > 120 && pose_counter < 160) /*|| (pose_counter > 600 && pose_counter < 700)*/ /*)*/ ){
    // noiseModel::Diagonal::shared_ptr unaryNoise = noiseModel::Diagonal::Sigmas(Vector3(0.1, 0.1, 0.5)); // 10cm std on x,y
    // graph.emplace_shared<UnaryFactor>(X(pose_counter), north, east, unaryNoise);
    // graph.emplace_shared<GPSFactor>(X(pose_counter), Point3(north, east, 0), unaryNoise);
    // noiseModel::Diagonal::shared_ptr attitudeNoise = noiseModel::Diagonal::Sigmas(Vector3(0.1, 0.1, 0.1));
    Vector6 sigmas;
    sigmas << Vector3(0.1, 0.1, 1.0), Vector3(0.15, 0.15, 0.1);
    // Pose3 delta(Rot3::Rodrigues(0.1, 0.1, 0.05), Point3(0.1, 0.1, 0.5));
    auto noise = noiseModel::Diagonal::Sigmas(sigmas);
    // graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
    graph.push_back(PriorFactor<Pose3>(X(pose_counter), Pose3(Rot3::yaw(theta), Point3(north, east, 0)), noise));
    use_nav = true;
    // graph.emplace_shared<Pose3AttitudeFactor>(X(pose_counter), Rot3::yaw(theta), unaryNoise);
  } 

  // cout << "Pose from GPS: " << Pose3(Rot3::yaw(theta), Point3(north, east, 0)) << endl;

  // This needs the current transformation
  
  translation << pose_msg.pose.position.x, pose_msg.pose.position.y, pose_msg.pose.position.z;

  q.x() = pose_msg.pose.orientation.x;
  q.y() = pose_msg.pose.orientation.y;
  q.z() = pose_msg.pose.orientation.z;
  q.w() = pose_msg.pose.orientation.w;
  
  rotation = q.toRotationMatrix();

  // std::cout << "Before: \n" << total_rotation << std::endl;
  // std::cout << "Before: \n" << total_translation << std::endl;
  
  total_rotation = total_rotation * rotation; 
  total_translation = total_translation + total_rotation*translation;
  
  // std::cout << "Recieved: \n" << rotation << std::endl;
  // std::cout << "Recieved: \n" << translation << std::endl;

  // std::cout << "Total: \n" << total_rotation << std::endl;
  // std::cout << "Total: \n" << total_translation << std::endl;

  Rot3 rot;
  Point3 point;
  
  if (use_nav){
    // total_pose = Pose3((Rot3::yaw(theta), Point3(north, east, 0)));
    rot = Rot3::yaw(theta);
    point = Point3(north, east, 0);
    // cout << "Just to make sure" << endl;
  } else{
    rot = (Rot3(total_rotation(0,0), total_rotation(0,1), total_rotation(0,2), total_rotation(1,0), total_rotation(1,1), total_rotation(1,2), total_rotation(2,0), total_rotation(2,1), total_rotation(2,2)));
    point = Point3(total_translation(0), total_translation(1), total_translation(2));

    // Vector6 sigmas;
    // sigmas << Vector3(0.3, 0.3, 0.3);
    // sigmas << Vector3(0.3, 0.3, 0.3);
    // auto rotation_noise = noiseModel::Diagonal::Sigmas(Vector3(0.01, 0.01, 0.4));
    // PoseRotationPrior<Pose3> factor(X(pose_counter), Rot3::yaw(theta_orientation), rotation_noise);
    // graph.push_back(factor);
    // graph.push_back(PoseRotationPrior<Pose3>(X(pose_counter), Pose3(Rot3::yaw(theta_orientation), point), rotation_noise));
    // graph.push_back(PoseRotationPrior<Rot3>(X(pose_counter), Rot3::yaw(theta_orientation), rotation_noise));

    // rot = (Rot3::Yaw(acos(total_rotation(0,0))));
    // point = Point3(total_translation(0), total_translation(1), 0);
    // total_pose = Pose3((Rot3(total_rotation(0,0), total_rotation(0,1), total_rotation(0,2), total_rotation(1,0), total_rotation(1,1), total_rotation(1,2), total_rotation(2,0), total_rotation(2,1), total_rotation(2,2)), Point3(total_translation(0), total_translation(1), total_translation(2))));
  } 
  Pose3 total_pose(rot, point);

  // cout << pose_counter << " " << "Total pose: " << total_pose << endl;

  // cout << "GPS data: " << Rot3::yaw(theta) << Point3(north, east, 0) << endl;

  imu.propState = imu.preintegrated->predict(imu.prevState, imu.prevBias);

  // total_pose2 = total_pose;
  // This needs the total transformation
  // initialEstimate.insert(X(pose_counter), total_pose.compose(delta));
  initialEstimate.insert(X(pose_counter), imu.propState.pose().compose(delta));
  // initialEstimate.insert(X(pose_counter), imu.propState.pose());
  // initialEstimate.insert(X(pose_counter), total_pose);
  initialEstimate.insert(V(pose_counter), imu.propState.velocity());
  // initialEstimate.insert(V(pose_counter), Vector3(0,0,0));
  initialEstimate.insert(B(pose_counter), imu.prevBias);

  CombinedImuFactor imuFactor(X(pose_counter - 1), V(pose_counter - 1),
                              X(pose_counter), V(pose_counter), B(pose_counter - 1),
                              B(pose_counter), *imu.preintegrated);

  // graph.add(imuFactor);

  // VelocityConstraint velConst(X(pose_counter - 1), X(pose_counter), 2);
  // graph.add(velConst);

  // do_more();

  ISAM2Result result = isam.update(graph, initialEstimate);
  int numberOfUpdates = 50;
  for (int i = 0 ; i < numberOfUpdates; ++i){
    isam.update();
  }
  // result.print();
  // ofstream os("Testing.dot");
  // const std::string os = "Testing.dot";
  // Values currentEstimate = isam.calculateEstimate();
  Values currentEstimate = isam.calculateBestEstimate();


  // imu.propState = imu.preintegrated->predict(imu.prevState, imu.prevBias);
  // imu.prevState = NavState(currentEstimate.at<Pose3>(X(lastFrame)),
  //                           currentEstimate.at<Vector3>(V(lastFrame)));
  // imu.prevBias = currentEstimate.at<imuBias::ConstantBias>(B(lastFrame));
  // imu.preintegrated->resetIntegrationAndSetBias(imu.prevBias);

  // std::cout << "Landmark number: " << landmark_counter << ", pose counter: " << pose_counter <<  ", x: " << currentEstimate.at<Pose3>(X(pose_counter)).x() << ", y: " << currentEstimate.at<Pose3>(X(pose_counter)).y() << ", z: " << currentEstimate.at<Pose3>(X(pose_counter)).z() << std::endl;
  // std::cout << "Speed or superspeed? " << currentEstimate.at<Pose3>(V(pose_counter)).x() << ", " << currentEstimate.at<Pose3>(X(pose_counter)).y() << endl;
  // cout << "Bias: " << imu.prevBias << endl;
  // std::cout << currentEstimate.at<Vector3>(V(pose_counter)) << std::endl;
  // cout << "Marginal covariance: \n" << isam.marginalCovariance(X(pose_counter)) << endl;

  total_pose2 = currentEstimate.at<Pose3>(X(pose_counter));

  // cout << "Total pose 2: " << total_pose2 << endl;

  Matrix3 rotate_after = total_pose2.rotation().matrix();

  total_rotation << rotate_after(0), rotate_after(1), rotate_after(2), rotate_after(3), rotate_after(4), rotate_after(5), rotate_after(6), rotate_after(7), rotate_after(8);
  total_translation << total_pose2.x(), total_pose2.y(), total_pose2.z();
  // total_rotation << rotate_after(0), rotate_after(1), 0, rotate_after(3), rotate_after(4), 0, 0, 0, rotate_after(8);
  // total_translation << total_pose2.x(), total_pose2.y(), 0;

  // std::cout << "After: \n" << total_rotation << std::endl;
  // std::cout << "After: \n" << total_translation << std::endl;
  if (pose_counter == 10){
    isam.saveGraph("First10.dot");
  }

  // graph.saveGraph(os, currentEstimate);

  ofstream myfile;
  myfile.open ("pose_estimates.txt");
  for (size_t i = 0; i < pose_counter + 1; ++i)
  {
    myfile << currentEstimate.at<Pose3>(X(i)).x() << ", " << currentEstimate.at<Pose3>(X(i)).y() << ", " << currentEstimate.at<Pose3>(X(i)).z() << "\n";
  } 
  myfile.close();

  ofstream myfile_;
  myfile_.open ("landmark_estimates.txt");
  for (size_t i = 1; i < number_landmarks + 1; ++i)
  {
    if (currentEstimate.exists(L(i))){
      myfile_ << currentEstimate.at<Point3>(L(i)).x() << " " << currentEstimate.at<Point3>(L(i)).y() << " " << currentEstimate.at<Point3>(L(i)).z() << "\n";
    }
  } 
  myfile_.close();


  if (nav_online){
    file << north << ", " << east << ", " << 0 << endl;
  }

  ofstream covariance_file;
  covariance_file.open ("covariance_estimates.txt");
  for (size_t i = 1; i < pose_counter + 1; ++i)
  {
    covariance_file << isam.marginalCovariance(X(i)) << "\n" << endl;
  } 
  covariance_file.close();
  // cout << isam.marginalCovariance(X(pose_counter)) << endl;

  // cout << "For pose number: " << pose_counter << "\n" << endl;

  // for (size_t i = 0; i < pose_counter + 1; ++i)
  // {
  //   cout << currentEstimate.at<Pose3>(X(i)).x() << ", " << currentEstimate.at<Pose3>(X(i)).y() << ", " << currentEstimate.at<Pose3>(X(i)).z() << "\n";
  // } 
  // file.close();

  std::cout << "Landmark number: " << landmark_counter << ", pose counter: " << pose_counter <<  ", x: " << currentEstimate.at<Pose3>(X(pose_counter)).x() << ", y: " << currentEstimate.at<Pose3>(X(pose_counter)).y() << ", z: " << currentEstimate.at<Pose3>(X(pose_counter)).z() << std::endl;

  // std::cout << currentEstimate.at<Vector3>(V(pose_counter)) << std::endl;

  // Eigen::Quaternionf quat = currentEstimate.at<Pose3>(X(pose_counter)).rotation;

  // geometry_msgs::PoseStamped pose_msg2;
  // geometry_msgs::Pose pose2; 
  // pose_msg2.header.frame_id = pose_msg.header.frame_id;
  // pose_msg2.header.stamp = ros::Time::now();//point_cloud_ptr->header.stamp;
  
  // pose2.position.x = currentEstimate.at<Pose3>(X(pose_counter)).x;
  // pose2.position.y = currentEstimate.at<Pose3>(X(pose_counter)).y;
  // pose2.position.z = currentEstimate.at<Pose3>(X(pose_counter)).z;

  // pose2.orientation = quat;

  // pose_msg2.pose = pose2;

  // pub2.publish(pose_msg2);

  imu.prevState = NavState(currentEstimate.at<Pose3>(X(pose_counter)),
                            currentEstimate.at<Vector3>(V(pose_counter)));
  imu.prevBias = currentEstimate.at<imuBias::ConstantBias>(B(pose_counter));
  imu.preintegrated->resetIntegrationAndSetBias(imu.prevBias);

  graph.resize(0);
  initialEstimate.clear();
  nav_online = false;
}

void do_more(){
  // lastFrame = pose_counter;
  if (lastFrame==1){
    auto priorPoseNoise = noiseModel::Diagonal::Sigmas(
    (Vector(6) << Vector3::Constant(0.1), Vector3::Constant(0.2)).finished());
    Pose3 initial(Rot3(cos(0.83), sin(0.83), 0, -sin(0.83), cos(0.83), 0, 0, 0, 1), Point3(-0.645953733353, 2.79517319837, 0));
    graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
    initialEstimate.insert(X(0), initial);
    graph.add(PriorFactor<imuBias::ConstantBias>(B(0), imu.priorImuBias,
                                               imu.biasNoiseModel));
    initialEstimate.insert(B(0), imu.priorImuBias);

    // Velocity prior - assume stationary
    graph.add(
        PriorFactor<Vector3>(V(0), Vector3(0, 0, 0), imu.velocityNoiseModel));
    initialEstimate.insert(V(0), Vector3(0, 0, 0));
  }

  if (lastFrame % 10 == 0 || lastFrame < 10){
    // noiseModel::Diagonal::shared_ptr correction_noise = noiseModel::Isotropic::Sigma(3,0.1);
    // GPSFactor gps_factor(X(lastFrame),
    //                       Point3(north,  // N,
    //                             east,  // E,
    //                             0), // D,
    //                       correction_noise);
    // graph.add(gps_factor);
    Vector6 sigmas;
    sigmas << Vector3(0.1, 0.1, 0.05), Vector3(0.1, 0.1, 0.5);
    Pose3 delta(Rot3::Rodrigues(0.1, 0.1, 0.05), Point3(0.1, 0.1, 0.5));
    auto noise = noiseModel::Diagonal::Sigmas(sigmas);
    // graph.push_back(PriorFactor<Pose3>(X(0), initial, priorPoseNoise));
    graph.push_back(PriorFactor<Pose3>(X(lastFrame), Pose3(Rot3::yaw(theta), Point3(north, east, 0)), noise));
  }
  // if (lastFrame == 200){
  //   cout << imu.prevBias << endl;
  // }

  imu.propState = imu.preintegrated->predict(imu.prevState, imu.prevBias);

  initialEstimate.insert(X(lastFrame), imu.prevState.pose());
  initialEstimate.insert(V(lastFrame), imu.prevState.velocity());
  // initialEstimate.insert(V(lastFrame), Vector3(0.0, 0.0, 0.0));
  initialEstimate.insert(B(lastFrame), imu.prevBias);

  CombinedImuFactor imuFactor(X(lastFrame - 1), V(lastFrame - 1),
                              X(lastFrame), V(lastFrame), 
                              B(lastFrame - 1), B(lastFrame), *imu.preintegrated);

  graph.add(imuFactor);
  // imuBias::ConstantBias zero_bias(Vector3(0, 0, 0), Vector3(0, 0, 0));
  // graph.add(BetweenFactor<imuBias::ConstantBias>(B(lastFrame-1),
  //                                                 B(lastFrame  ),
  //                                                 zero_bias, imu.biasNoiseModel));

  isam.update(graph, initialEstimate);

  for (int i = 0 ; i < 10; ++i){
    isam.update();
  }

  Values currentEstimate = isam.calculateBestEstimate();

  imu.prevState = NavState(currentEstimate.at<Pose3>(X(lastFrame)),
                            currentEstimate.at<Vector3>(V(lastFrame)));
  imu.prevBias = currentEstimate.at<imuBias::ConstantBias>(B(lastFrame));
  imu.preintegrated->resetIntegrationAndSetBias(imu.prevBias);

  graph.resize(0);
  initialEstimate.clear();

  std::cout << "x: " << currentEstimate.at<Pose3>(X(lastFrame)).x() << ", y: " << currentEstimate.at<Pose3>(X(lastFrame)).y() << ", z: " << currentEstimate.at<Pose3>(X(lastFrame)).z() << std::endl;
  std::cout << currentEstimate.at<Vector3>(V(lastFrame)) << std::endl;
  cout << "GPS: " << north << ", " << east << endl;
  cout << "Bias: " << imu.prevBias << endl;



  lastFrame++;
}

void do_imu(const sensor_msgs::ImuConstPtr& imu_msg){
  // cout << "Angular velocity x: " << imu_msg->angular_velocity.x << " y: " << imu_msg->angular_velocity.y << " z: " << imu_msg->angular_velocity.z << endl;
  // cout << "Linear acceleration x: " << imu_msg->linear_acceleration.x << " y: " << imu_msg->linear_acceleration.y << " z: " << imu_msg->linear_acceleration.z << endl;

  double dt = 0.01;
  Vector3 gyr(imu_msg->angular_velocity.x, imu_msg->angular_velocity.y, imu_msg->angular_velocity.z);
  Vector3 acc(imu_msg->linear_acceleration.x, imu_msg->linear_acceleration.y, imu_msg->linear_acceleration.z);

  // double pitch = atan2(imu_msg->angular_velocity.x, sqrt(imu_msg->angular_velocity.y*imu_msg->angular_velocity.y+imu_msg->angular_velocity.z*imu_msg->angular_velocity.z));
  // double roll = atan2(imu_msg->angular_velocity.y, sqrt(imu_msg->angular_velocity.x*imu_msg->angular_velocity.x+imu_msg->angular_velocity.z*imu_msg->angular_velocity.z));
  // cout << "Pitch angle: " << pitch << ", roll: " << roll << endl;

  // gyr = gyr;
  // acc = orientation * acc;

  imu.preintegrated->integrateMeasurement(acc, gyr, dt);

  imu_counter++;

  // if (imu_counter % 200 == 0){
  //   do_more();
  // }
}

void ori_test(const isam2::orientationEstimate& orientation_msg){
  // -47.6 deg vs -78.5 deg
  double yaw = orientation_msg.yaw+126.6;
  if (yaw > 180){
    yaw = yaw - 360;
  } else if (yaw < -180){
    yaw = yaw + 360;
  }
  // cout << "Roll: " << 3.14/180*orientation_msg.roll << ", pitch: " << 3.14/180*orientation_msg.pitch << ", yaw: " << 3.14/180*-yaw << endl;
  theta_orientation = -yaw*3.14/180;
}

void navi_test(const isam2::NorthEastHeading& nav_msg){
  north = nav_msg.north;
  east = - nav_msg.east;
  theta = - nav_msg.heading;
  nav_online = true;
  // cout << Rot3::yaw(theta) << endl;
}


int main(int argc, char* argv[]) {

  ros::init(argc, argv, "isam2");
  ros::NodeHandle nh;
  
  std::cout << "Starting the graph shit" << std::endl;

  // parameters.relinearizeThreshold = 0.01;
  // parameters.relinearizeSkip = 1;
  // parameters.cacheLinearizedFactors = false;
  // parameters.enableDetailedResults = true;
  // parameters.optimizationParams = ISAM2DoglegParams();
  ISAM2 isam(parameters);

  file.open("GPS");

  total_rotation << cos(0.83), sin(0.83), 0, -sin(0.83), cos(0.83), 0, 0, 0, 1;
  total_translation << -0.645953733353, 2.79517319837, 0;

  // auto priorPoseNoise = noiseModel::Diagonal::Sigmas(
  //     (Vector(6) << Vector3::Constant(0.1), Vector3::Constant(0.1)).finished());
  // graph.emplace_shared<PriorFactor<Pose3>>(X(0), Pose3::identity(),
  //                                          priorPoseNoise);
  // initialEstimate.insert(X(0), Pose3::identity());

  // // Bias prior
  // graph.add(PriorFactor<imuBias::ConstantBias>(B(0), imu.priorImuBias,
  //                                              imu.biasNoiseModel));
  // initialEstimate.insert(B(0), imu.priorImuBias);

  // // Velocity prior - assume stationary
  // graph.add(
  //     PriorFactor<Vector3>(V(0), Vector3(0, 0, 0), imu.velocityNoiseModel));
  // initialEstimate.insert(V(0), Vector3(0, 0, 0));
  
  ros::Subscriber sub = nh.subscribe("/xsens/imu", 1, do_imu);

  // ros::Subscriber sub2 = nh.subscribe("/Matching_points", 1, print_shit);

  ros::Subscriber sub3 = nh.subscribe("/Point_test", 1, point_test);

  ros::Subscriber sub4 = nh.subscribe("/Pose_test", 1, pose_test);

  ros::Subscriber sub5 = nh.subscribe("/xsens/orientation", 1, ori_test);

  ros::Subscriber sub6 = nh.subscribe("/navigation/eta", 1, navi_test);

  pub = nh.advertise<sensor_msgs::PointCloud2> ("Testing_this", 1);

  pub2 = nh.advertise<geometry_msgs::PoseStamped> ("Correct_pose",1);

  ros::spin();

  return 0;
}